# ウディタ用 単体テストフレームワーク

## バージョン履歴
* V1.0:2020/5/10:新規作成

## はじめに

ウディタを使っていてこんな時はありませんか？
* 複数のコモンイベントにまたがったイベントを作成したがうまく動作しない。原因がどのコモンイベントにあるか分からない
* コモンイベントを作ったけど手動での動作確認がとても手間
* 作ったコモンイベントの動作や書き方を見直したいけど、バグってしまうのが怖くて触れない

単体テストを作ってこんな悩みとはオサラバしましょう。  
この「単体テストフレームワーク」では、ウディタのコモンイベントの単体テストを作ることを補助する以下のものを同梱しています。
* 単体テスト.common
  * 単体テストをサポートするコモンイベント群からなるコモンファイルです。
  * コモンイベント11個分の連続したスペースが必要です。(並び変え・名称変更は一部不可)
  * DB、変数は不要です。
* UnitTestTemplate.mps
  * 単体テストをマップイベントで作成する場合のテンプレートです。
* TestTourTemplate.mps
  * 複数のマップイベントで作られた単体テスト群をまとめて実行するマップイベントを作成する場合のテンプレートです。
* Data/フォルダ以下ファイル一式
  * 単体テストフレームワークを実際に使ってテストしているサンプルです。
  * 単体テストフレームワーク自体の単体テストも含まれています。

## 単体テスト実装の流れ

ここでは、単体テストフレームワークを使ったオススメの単体テストの方法の全体の流れを説明します。  
(後述)とある点については、「コモンファイルマニュアル」の節を確認してください。

1. 「TestTourTemplate.mps」をコピーして、ウディタでマップを読み込みます。任意の名前でマップを登録します。
    * 「単体テスト」という名前が分かりやすいです。
    * mpsファイルは「Data」フォルダ以下に「UnitTest」フォルダを作成して格納しておくとよいでしょう。
2. 「UnitTestTemplate.mps」をコピーして、ウディタでマップを読み込みます。任意の名前でマップを登録します。
    * 1つのテスト対象のコモンイベントごとに、1つのマップを作成します。
    * テスト対象となるコモンイベントの名前を付けると分かりやすいです。
    * 1.で作成したマップの子に移動しておくとよいでしょう。
3. 2.のマップに配置されているイベントを編集して、単体テストコードを実装します。(後述)
4. 1.のマップに配置されているイベントを編集して、3.のマップに移動できるようにします。(後述)
5. 単体テストを実行したい場合、1.のマップを「ゲームの開始位置」に設定し、テストプレイを実行します。
    * 単体テストが自動で実行されます。テストNGがある場合、画面上にNG内容がメッセージとして出力されます。

以降では、2.で作成したマップイベントを「単体テスト」、1.で作成したマップイベントを「テストツアー」と呼びます。
* 単体テスト
  * コモンイベントを呼び出し、期待値と戻り値が一致するかを照合するイベントのこと。
* テストツアー
  * 複数の単体テストを一括で実行する、とりまとめ用のイベントのこと。

## コモンファイルマニュアル

「単体テスト.common」に含まれる11個のコモンイベントのマニュアルです。
* ★が「単体テスト」を実装する上で重要なコモンイベントです。
* ☆が「テストツアー」を実装する上で重要なコモンイベントです。

### テストメッセージ出力
このコモンイベントは他のコモンイベントから呼び出されます。  
単体テストを実装する開発者が呼び出すことは基本的にありません。

* 引数・戻り値
  * 省略

### ★単体テスト対象設定
単体テストを実施したいコモンイベントのIDを指定することで、単体テストに関する事前設定を実行します。  
必ず「単体テストケース設定」「単体テスト実行」より前に呼び出してください。

単体テストを実施したいコモンイベントの位置を変更した場合、ここで設定したIDを書き換える必要があります。  
コモンイベントを名前により指定することは出来ません。

* 引数
  * 0 : コモンイベントID : テスト対象となるコモンイベントのID
* 戻り値
  * なし

### ★単体テストケース設定
次に実行する単体テストでの期待値を指定します。  
必ず「単体テスト対象設定」よりも後、「単体テスト実行」よりも前に呼び出してください。  
期待値は、「戻り値変数番号」と「戻り値」の組で指定します。  
ここで指定する「戻り値変数番号」は、テスト対象のコモンイベントの「結果を返す値」として指定した変数の番号を指定することを想定していますが、それ以外の変数番号も指定することが出来ます。

* 引数
  * 0 : 戻り値変数番号 : コモンイベントの実行完了後に値を確認したいコモンイベント変数の変数番号
  * 1 : 戻り値数値 : テスト対象を実行した後に期待する戻り値
  * 2 : 設定モード : テストケースの設定方法を変更するための引数
    * 0 : 上書き : 基本的にはこの初期値から変更不要です。
    * 1 : 期待値の追加 : 期待値を2個以上に増やしたい場合、2回目以降の呼び出しはこのモードにしてください。
    * 2 : 初期化 : 呼び出し不要です。
* 戻り値
  * なし

### ★単体テスト実行
テスト対象を実行し、「単体テストケース設定」で設定された期待値と照合します。  
期待値と実際の値が異なるときは、その旨をメッセージで出力します。  
必ず「単体テスト対象設定」「単体テストケース設定」より後に呼び出してください。

* 引数
  * 0～3 : 引数1～4 : テスト対象のコモンイベントに渡す数値型引数の値です。
    * テスト対象のコモンイベントで「入力する値の意味」を設定していない引数は、0のまま変更不要です。
  * 5～8 : 引数5～8 : テスト対象のコモンイベントに渡す文字列型引数の値です。
* 戻り値
  * なし

### ☆単体テストツアー
単体テストが連番のマップの自動イベントで実装されているとき、それらを順に実行します。  
必ず以下のように実行してください。
1. マップイベントから呼び出してください。
2. 呼び出し元のマップイベントでは、呼出し後にページが切り替わるようにセルフ変数を設定してください。
    * テストツアーの移動先の自動イベントより呼出し元のイベント処理が優先されるため、ページを切り替えないと無限ループが発生します。
3. 実行する単体テストのマップIDで、自動実行のイベントを実装してください。
    * テストツアーを実行すると、該当マップへ移動し、その自動実行イベントが次に呼び出されます。
4. 単体テストの自動実行イベントの最後に、「単体テストツアー復帰」コモンイベントを呼び出してください。

上記のようにすることで、
```
[1.のマップイベント]
→[本イベント]→[3.のマップイベント]
→[本イベント]→[3.のマップイベント]
→…
→[1.のマップイベント]
```
のような順に主人公が場所移動し、設定された自動イベントが次々に実行されます。

* 引数
  * 0 : 開始マップID : 実行したい単体テストのマップIDの連番の開始となるIDを指定してください。
  * 1 : 終了マップID : 実行したい単体テストのマップIDの連番の最後となるIDを指定してください。
    * 1つのテストだけ実行する場合は、終了マップID=開始マップIDとなるよう指定してください。
  * 2 : 継続か : 設定不要です。
* 戻り値
  * なし

### ☆単体テストツアー復帰
単体テストのマップイベントで処理が終了した場合に呼び出してください。  
テストツアーの連続実行処理に戻ることが出来ます。

* 引数
  * なし
* 戻り値
  * なし

### 数値検証
数値を検証するユーティリティです。  
２つの値を比較して不一致の場合はメッセージを出力します。

* 引数・戻り値
  * 省略

### 文字列検証
文字列を検証するユーティリティです。  
２つの値を比較して不一致の場合はメッセージを出力します。

* 引数・戻り値
  * 省略

### 可変DB値検証
可変DBの値を検証するユーティリティです。  
指定したDBタイプ番号・データ番号・項目番号での可変DBの値と期待値を比較して不一致の場合はメッセージを出力します。

* 引数・戻り値
  * 省略

### メッセージ検証
メッセージが表示されているかどうかを検証するユーティリティです。  
検証したいコモンイベントの呼び出し直後に本コモンイベントを呼び出してください。  
また、連続で処理を実行している場合、かつ、メッセージが表示されていないことを検証する場合、テスト対象のコモンイベント呼び出しの直前に1フレームのウェイトを挿入してください。  
「メッセージ表示中か」システム変数を参照しますが、ウェイトがないと変数の値が更新されていない可能性があるためです。

* 引数
  * 0 : 検証動作 : 検証内容を指定します。
    * 0 : 表示なし : メッセージが表示されていないことを検証します。
    * 1 : 表示あり : メッセージが表示されていることを検証します。メッセージの中身については検証しません。
    * 2 : 表示文字列も検証 : メッセージが表示されていることを検証します。さらにそのメッセージの中身が「期待文字列」に一致することも検証します。
  * 5 : 期待文字列 : メッセージ文字列の内容の期待値です。
    * 検証動作が2のときだけ使用されます。
* 戻り値
  * なし

### メッセージ自動送り
テスト実行中に表示されたメッセージを自動で閉じることで、テストをキー操作なしで終了させることができるユーティリティです。
このコモンイベントは並列実行(常時)により動作します。
このコモンイベントの動作を無効にするには、以下のいずれかを行ってください。
* 「動作モード」を「停止」にして呼び出す
* 「動作モード」を「開始」している呼び出しを削除する
* このコモンイベントの起動条件を「呼び出しのみ」に変更する

このコモンイベントが実行中の場合は、単体テストフレームワークのもつコモンイベントの呼び出しが不正だった場合のエラーメッセージも自動で閉じます。  
単体テストを実装中、単体テストの動作を確認する場合は、このコモンイベントの動作を無効にすることをオススメします。  
なお、「テストメッセージ出力」で表示した文字列は、自動では閉じません。

* 引数
  * 0 : 動作モード : 動作を変更します。
    * 0 : 停止 : 停止させます
    * 1 : 実行 : 実行させます

## テンプレートマニュアル

### 単体テストテンプレート
同梱する「UnitTestTemplate.mps」では、以下のマップイベントを実装しています。
1. ページ1を、自動実行します。
2. 不正なIDを指定した「単体テスト対象設定」を呼び出します。
3. 「単体テストケース設定」を初期値で呼び出します。
4. 「単体テスト実行」をすべての引数を0で呼び出します。
5.すべての処理が終わったらテストツアーに復帰します。

特に何も変更しない場合、2.および4.でエラーメッセージが表示(※)されます。その後、テストツアーに処理が戻ります。  
単体テストの処理内容を実装する場合には、「コモンファイルマニュアル」の★の付いたメソッドの引数を正しく指定してください。

※ただし、テストツアーテンプレートを使っている場合、メッセージが自動送りされるため、エラーメッセージを目視確認することはできません。

### テストツアーテンプレート
同梱する「TestTourTemplate.mps」では、以下のマップイベントを実装しています。
1. ページ1～10をもち、すべて自動実行します。
2. ページ1～9では何もしない「単体テストツアー」を実行し、次のページに遷移します。
3. ページ10では、このイベントを消去します。

上記1.～3.により、特に何も変更をくわえない場合、ページ1～10を巡った後に、イベントが終了し、テストツアーのマップに主人公がとどまります。

#### 単体テストを実行する場合
ページ1～9に実装されている「単体テストツアー」の「開始マップID」「終了マップID」を書き換えて、テスト実行したい単体テストのマップを指定してください。

#### 単体テスト実行後の動作を変更する場合
例えばテストプレイの開始前にすべての単体テストを実行しておくと便利です。  
この場合、ページ10に、テストプレイで手動操作したい場所へ移動する処理を実装してください。

#### テスト時のメッセージ送りを制御する場合
テストツアーテンプレートで作成したテストツアーは、テスト時のメッセージ送りを自動で閉じます。  
この挙動を変更したい場合は、「コモンファイルマニュアル」の「メッセージ自動送り」の節を確認ください。

## サンプルの確認方法
1. ウディタ一式の複製を用意してください。（基本システム不要です）
2. Dataフォルダをペーストしてください。（重複するファイルは上書き）
3. Editor.exeの「サンプル」マップで内容を確認してください。

## 免責
このコモンイベントを使用したことによる一切の責任を作者は負いません。
使用にあたっては自己責任でお願いします。

## 作者

千字肇（せんじけい）  
Twitter: @1000ji_K   
(質問はこちらのDMで随時受け付けます。レスポンスの良さは期待しないでください。)
